import mozanalysis.bayesian_stats.binary as mabsbin
import pandas as pd
import scipy.stats as st
import numpy as np


def run_simmo(fake_results, control_rate, true_rel_change):
    lower = []
    upper = []

    for _ in range(1000):
        fake_results['num_conversions'] = st.binom(
            fake_results['num_enrollments'],
            np.array([control_rate, control_rate * (1 + true_rel_change)])
        ).rvs()

        stats = mabsbin.compare_branches_from_agg(fake_results)

        lower.append(stats['comparative']['test']['rel_uplift']['0.975'])
        upper.append(stats['comparative']['test']['rel_uplift']['0.025'])

    return pd.DataFrame({'lower': lower, 'upper': upper}).quantile([0.025, 0.975])


def set_up_fake_results_df(total_num_enrollments, test_branch_prop):
    t_pop = st.binom(total_num_enrollments, test_branch_prop).rvs()

    return pd.DataFrame(
        {'num_enrollments': [total_num_enrollments - t_pop, t_pop]},
        index=['control', 'test']
    )


def vary_rel_change(
    total_num_enrollments, test_branch_prop, control_rate, rel_change_list, ax,
    metric_name
):
    """Plot how various relative changes are likely to be reported.

    Can usually get away with only two `rel_change_list`; linear
    interpolation is great.

    Example usage in databricks:

        plt.close()
        plt.clf()
        ax = plt.figure().gca()

        vary_rel_change(100000, 0.5, 0.1, [0, 0.2], ax, 'fake metric')

        ax.figure.tight_layout()
        ax.figure.show()
        display()
    """
    fake_results = set_up_fake_results_df(total_num_enrollments, test_branch_prop)

    res = {rc: run_simmo(fake_results, control_rate, rc) for rc in rel_change_list}

    ax.fill_between(
        rel_change_list,
        [res[rc].loc[0.025, 'lower'] for rc in rel_change_list],
        [res[rc].loc[0.975, 'lower'] for rc in rel_change_list],
        alpha=0.1,
        color='red',
        label="2.5% worst case"
    )
    ax.fill_between(
        rel_change_list,
        [res[rc].loc[0.025, 'upper'] for rc in rel_change_list],
        [res[rc].loc[0.975, 'upper'] for rc in rel_change_list],
        alpha=0.1,
        color='green',
        label="97.5% best case"
    )

    for rc in rel_change_list:
        ax.plot([rc, rc], res[rc]['lower'].values, 'red')
        ax.plot([rc, rc], res[rc]['upper'].values, 'green')

    ax.plot([min(rel_change_list), max(rel_change_list)], [0, 0], 'k--', label='zero')
    ax.legend()
    ax.set_xlabel("'true' relative change w.r.t. control")
    ax.set_ylabel('reported relative change (95% CI)')
    ax.set_title(
        'Expected range of best/worst cases:\n'
        + metric_name + '\n'
        + '{e} enrollments, {t}% test branch, {r} baseline rate'.format(
            e=total_num_enrollments, t=test_branch_prop*100, r=control_rate
        )
    )


def vary_population_size(
    total_num_enrollments_list, test_branch_prop, control_rate, rel_change, ax,
    metric_name
):
    """Plot how various population sizes are likely to be reported.

    E.g. Find how large a population is required to reliably observe an
    effect as being positive.

    Linear interpolation is poor; supply several population sizes in
    `total_num_enrollments_list`.

    Example usage in databricks:

        plt.close()
        plt.clf()
        ax = plt.figure().gca()

        vary_population_size(
            [10000, 50000, 100000], 0.5, 0.1, 0.05, ax, 'fake metric'
        )

        ax.figure.tight_layout()
        ax.figure.show()
        display()
    """

    res = {
        tne: run_simmo(
            set_up_fake_results_df(tne, test_branch_prop),
            control_rate,
            rel_change
        )
        for tne in total_num_enrollments_list
    }

    ax.fill_between(
        total_num_enrollments_list,
        [res[rc].loc[0.025, 'lower'] for rc in total_num_enrollments_list],
        [res[rc].loc[0.975, 'lower'] for rc in total_num_enrollments_list],
        alpha=0.1,
        color='red',
        label="2.5% worst case"
    )
    ax.fill_between(
        total_num_enrollments_list,
        [res[rc].loc[0.025, 'upper'] for rc in total_num_enrollments_list],
        [res[rc].loc[0.975, 'upper'] for rc in total_num_enrollments_list],
        alpha=0.1,
        color='green',
        label="97.5% best case"
    )

    for rc in total_num_enrollments_list:
        ax.plot([rc, rc], res[rc]['lower'].values, 'red')
        ax.plot([rc, rc], res[rc]['upper'].values, 'green')

    ax.plot(
        [min(total_num_enrollments_list), max(total_num_enrollments_list)],
        [0, 0], 'k--', label='zero'
    )
    ax.legend()
    ax.set_xlabel("total population size")
    ax.set_ylabel('reported relative change (95% CI)')
    ax.set_title(
        'Expected range of best/worst cases:\n'
        + metric_name + '\n'
        + '{t}% test branch, {r} baseline rate, {rc} true rel change'.format(
            t=test_branch_prop*100, r=control_rate, rc=rel_change
        )
    )



# experiment_membership_df(slug, date_start, enrollment_period, observation_period)
# ms_pings_subset_df(df, date_start, total_period, columns=MS_USAGE_COLS, slug=None)
# as_pings_subset_df(as_df, date_start, total_period, slug=None)
# experiment_pings_df(pings_df, membership_df, observation_period)
# daily_usage_df(df, agg_functions, client_fields=client_fields)
# agg functions:
#   all:
#       - client_fields
#   main summary:
#       - daily_usage_aggs
#   as health:
#       - as_health_aggs
#   as sessions:
#       - as_session_aggs
#   as clicks:
#       - as_clicks_aggs
#   as snippets:
#       - as_snippets_aggs = get_as_snippets_aggs(message_id=None)

# spark imports
from pyspark.sql import SparkSession
import pyspark.sql.functions as F
from pyspark.sql.functions import udf
from pyspark.sql.types import MapType, StringType, BooleanType

# local imports
from .constants import MS_USAGE_COLS
from .activitystream import as_experiment_field, as_pref_setting
from .activitystream import as_health_default_homepage, as_health_default_newtab
from .util import date_plus_N, date_to_string, string_to_date


# fetch spark context
spark = SparkSession.builder.getOrCreate()


# ------------------- experiment membership ------------------

def experiment_membership_df(slug, date_start, enrollment_period, observation_period):
    """
    notes on this methodology:
    - enrollments and unenrollments: only distinct client_id, branch
      (multiple pings collapsed, first ping counted)
    - join unenrollments by client_id, branch
      (can have unenrollment in one branch and not in other)
    - unenrollments without branch are ignored
      (even if there is an enrollment ping with same client_id)

    table schema:
      client_id
      branch
      enrollment_dt
      ping_count_enrollment
      unenrollment_dt
      ping_count_unenrollment
    """

    # function can take datetime or the s3 date string
    if type(date_start) == str:
        date_start = string_to_date(date_start)

    # get dates: enrollment period and max obs period
    date_enroll_end = date_plus_N(date_start, enrollment_period)
    date_obs_end = date_plus_N(date_start, enrollment_period + observation_period)

    # convert everything into s3 date string format
    date_start = date_to_string(date_start)
    date_enroll_end = date_to_string(date_enroll_end)
    date_obs_end = date_to_string(date_obs_end)

    # prep events table
    events = spark.sql("select * from events")

    # ----------------- enrollment pings -----------------

    # get enrollments in the enrollment window for experiment
    enrollments = events.filter("event_category = 'normandy'")
    enrollments = enrollments.filter("event_method = 'enroll'")
    enrollments = enrollments.filter("submission_date_s3 >= '%s'" % date_start)
    enrollments = enrollments.filter("submission_date_s3 < '%s'" % date_enroll_end)
    enrollments = enrollments.filter("event_string_value = '%s'" % slug)

    # roll up to unique client_id, branch,
    #   if there are multiple pings, get the earliest enrollment.
    # also get the number of pings sent
    enrollments = enrollments.groupby([
        'client_id',
        'event_map_values.branch'
    ]).agg(
        F.min(F.col('submission_date_s3')).alias('submission_date_s3'),
        F.count('*').alias('ping_count_enrollment')
    )

    # rename columns for joining ease
    enrollments = enrollments.select([
        F.col('client_id').alias('client_id_enrollment'),
        F.col('branch').alias('branch_enrollment'),
        F.to_date(F.col('submission_date_s3'),
                  'yyyyMMdd').alias('first_branch_enrollment_dt'),
        F.col('ping_count_enrollment')
    ])

    # ------------ unenrollment pings --------------------

    # get unenrollments in the maximum observation period for experiment
    unenrollments = events.filter("event_category = 'normandy'")
    unenrollments = unenrollments.filter("event_method = 'unenroll'")
    unenrollments = unenrollments.filter("submission_date_s3 >= '%s'" % date_start)
    unenrollments = unenrollments.filter("submission_date_s3 <= '%s'" % date_obs_end)
    unenrollments = unenrollments.filter("event_string_value = '%s'" % slug)

    # group by client_id, branch. if there are multiple records, use the earliest date,
    # and get the number of pings sent
    unenrollments = unenrollments.groupby([
        'client_id',
        'event_map_values.branch'
    ]).agg(
        F.min(F.col('submission_date_s3')).alias('submission_date_s3'),
        F.count('*').alias('ping_count_unenrollment')
    )

    # rename columns for ease of joining
    unenrollments = unenrollments.select([
        F.col('client_id').alias('client_id_unenrollment'),
        F.col('branch').alias('branch_unenrollment'),
        F.to_date(F.col('submission_date_s3'),
                  'yyyyMMdd').alias('first_branch_unenrollment_dt'),
        F.col('ping_count_unenrollment')
    ])

    # ----------------- membership table ---------------

    # join the enrollment and unenrollment table
    # left side is the enrollments
    # the following unenrollments are included:
    #     client_id matches an enrollment
    #     branch matches an enrollment (if client_id exists
    #       but the branch doesn't match, it is ignored)
    #     unenrollment ping sent within observation period
    #       starting from enrollment date
    membership_table = enrollments.join(
        unenrollments,
        (F.col('client_id_enrollment') == F.col('client_id_unenrollment'))
        & (F.col('branch_enrollment') == F.col('branch_unenrollment'))
        & (F.datediff(F.col('first_branch_unenrollment_dt'),
                      F.col('first_branch_enrollment_dt')
                      ) < observation_period),
        how='left'
        )

    # rename columns to final form
    membership_table = membership_table.select([
        F.col('client_id_enrollment').alias('client_id'),
        F.col('branch_enrollment').alias('branch'),
        F.col('first_branch_enrollment_dt').alias('enrollment_dt'),
        F.col('ping_count_enrollment'),
        F.col('first_branch_unenrollment_dt').alias('unenrollment_dt'),
        F.col('ping_count_unenrollment'),
    ])

    return membership_table


# ------------------- pre-process pings ------------------

def ms_pings_subset_df(df, date_start, total_period, columns=MS_USAGE_COLS, slug=None):
    """
    get subset of main summary that covers period of interest and columns of interest
      note: will convert submission_date_s3 to a date type column (activity_dt)
      note: if slug is provided, will only return pings tagged with the experiment and
            also provide the branch as a column

    table schema
      client_id
      activity_dt
      branch (optional)
      usage cols
    """
    # function can take datetime or the s3 date string
    if type(date_start) == str:
        date_start = string_to_date(date_start)

    # get date end of maximum possible observation period
    date_obs_end = date_plus_N(date_start, total_period)

    # convert everything into s3 date string format
    date_start_str = date_to_string(date_start)
    date_obs_end_str = date_to_string(date_obs_end)

    # if we're looking back, switch start and end
    if date_obs_end < date_start:
        date_start_str = date_to_string(date_obs_end)
        date_obs_end_str = date_to_string(date_start)

    # ----------------- subset dates and columns -----------------

    df = df.filter("submission_date_s3 >= '%s'" % date_start_str)
    df = df.filter("submission_date_s3 <= '%s'" % date_obs_end_str)

    columns = ['client_id', 'submission_date_s3', 'experiments'] + columns
    df = df.select(columns)

    # ----------------- tagged only if slug provided -----------------

    if slug is not None:
        df = df.filter("experiments['%s'] is not null" % slug)
        df = df.withColumn('branch', F.col('experiments')[slug])

    # ----------------- clean up columns -----------------

    df = df.withColumn('activity_dt',
                       F.to_date(
                          F.col('submission_date_s3'),
                          'yyyyMMdd')
                       )
    df = df.drop("experiments")
    df = df.drop("submission_date_s3")

    return df


def as_pings_subset_df(as_df, date_start, total_period, slug=None):
    """
    get subset of activity stream pings with some columns standardized

    providing an experiment slug will add a branch field and remove pings
    without the slug

    table schema
      client_id
      activity_dt
      branch (optional)
      as cols

    """

    # function can take datetime or the s3 date string or as date string
    if type(date_start) == str:
        if '-' in date_start:
            date_start = string_to_date(date_start, '%Y-%m-%d')
        else:
            date_start = string_to_date(date_start)

    # get date end of maximum possible observation period
    date_obs_end = date_plus_N(date_start, total_period)

    # convert everything into as date string format
    date_start_str = date_to_string(date_start, '%Y-%m-%d')
    date_obs_end_str = date_to_string(date_obs_end, '%Y-%m-%d')

    # if we're looking back, switch start and end
    if date_obs_end < date_start:
        date_start_str = date_to_string(date_obs_end, '%Y-%m-%d')
        date_obs_end_str = date_to_string(date_start, '%Y-%m-%d')

    # ----------------- subset dates -----------------

    as_df = as_df.filter("date >= '%s'" % date_start_str)
    as_df = as_df.filter("date <= '%s'" % date_obs_end_str)

    # ----------------- tagged only if slug provided -----------------

    if slug is not None:
        # set up udf for parsing activity stream experiment field
        schema = MapType(StringType(), StringType())
        as_experiment_field_udf = udf(as_experiment_field, schema)

        # get experiments field into standard format
        as_df = as_df.withColumn('experiments',
                                 as_experiment_field_udf(F.col('shield_id'))
                                 )

        # keep only data tagged with experiment and get branch column
        as_df = as_df.filter("experiments['%s'] is not null" % slug)
        as_df = as_df.withColumn('branch', F.col('experiments')[slug])
        as_df = as_df.drop('experiments')

    as_df = as_df.withColumn('activity_dt', F.col('date'))
    as_df = as_df.drop('shield_id').drop('date')

    return as_df


# ------------------- keep only experiment pings ------------------

def experiment_pings_df(pings_df, membership_df, observation_period):
    """
    get the subset of pings that are:
        1: in the membership df (by client_id)
        2: activity date is after enrollment
           (at least the day after) and before end of observation period

    note: if the pings_df has a branch column, join on (client_id, branch)
    """
    # if observation period is negative, get the days preceding
    # if positive, get days following
    top = observation_period
    bottom = 0
    if observation_period < 0:
        top = 0
        bottom = observation_period

    # rename column so joining is easier
    pings_df = pings_df.withColumn(
        'client_id_pings',
        F.col('client_id')
                                  ).drop('client_id')

    # if branch exists, join on client_id, branch, date range
    if 'branch' in pings_df.columns:
        pings_df = pings_df.withColumn(
            'branch_pings',
            F.col('branch')
                                       ).drop('branch')
        df = membership_df.join(
            pings_df,
            (F.col('client_id') == F.col('client_id_pings'))
            & (F.col('branch') == F.col('branch_pings'))
            & (F.datediff(F.col('activity_dt'),
                          F.col('enrollment_dt')) < top)
            & (F.datediff(F.col('activity_dt'),
                          F.col('enrollment_dt')) > bottom),  # no enroll day pings
            how='left'
            )
        df = df.drop('branch_pings')

    else:
        df = membership_df.join(
            pings_df,
            (F.col('client_id') == F.col('client_id_pings'))
            & (F.datediff(F.col('activity_dt'),
                          F.col('enrollment_dt')) < top)
            & (F.datediff(F.col('activity_dt'),
                          F.col('enrollment_dt')) > bottom),  # no enroll day pings
            how='left')

    # clean up columns
    df = df.drop('client_id_pings')

    return df


# ------------------- client-daily rollups ------------------

client_fields = [
        F.last(F.col('enrollment_dt')).alias('enrollment_dt'),
        F.last(F.col('ping_count_enrollment')).alias('ping_count_enrollment'),
        F.last(F.col('unenrollment_dt')).alias('unenrollment_dt'),
        F.last(F.col('ping_count_unenrollment')).alias('ping_count_unenrollment')
    ]


# generic client-daily rollup function
def daily_usage_df(df, agg_functions, client_fields=client_fields):
    """
    generic function for rolling up experiment pings into a
    (client_id, branch, activity date) level
        note: the client_fields above are assumed to be provided by default
        note: you have to provide the agg functions and decide the roll up
              logic inside them individually
    """
    if client_fields:
        agg_functions = client_fields + agg_functions
    return df.groupby(['client_id', 'branch', 'activity_dt']).agg(*agg_functions)


# ------------------- client-daily agg functions ------------------

# agg functions:
#
# all:
#   client_fields
# main summary:
#   - daily_usage_aggs
# as health:
#   - as_health_aggs
# as sessions:
#   - as_session_aggs
# as clicks:
#   - as_clicks_aggs
# as snippets:
#   - as_snippets_aggs = get_as_snippets_aggs(message_id=None)


# agg fields: main summary usage ------------------

days_used = F.countDistinct(
                    F.col('activity_dt')).alias('days_used')

active_hours = F.sum(
                F.coalesce(
                    F.col('active_ticks'),
                    F.lit(0)
                          ) * F.lit(5) / F.lit(3600)).alias('active_hours')

uris = F.sum(
        F.coalesce(
            F.col('scalar_parent_browser_engagement_total_uri_count'),
            F.lit(0)
                  )).alias('uris')

tabs_opened = F.sum(
                F.coalesce(
                    F.col('scalar_parent_browser_engagement_tab_open_event_count'),
                    F.lit(0)
                          )).alias('tabs_opened')

windows_opened = F.sum(
                   F.coalesce(
                     F.col('scalar_parent_browser_engagement_window_open_event_count'),
                     F.lit(0)
                             )).alias('windows_opened')

ping_count = F.count(F.col('activity_dt')).alias('usage_ping_count')


daily_usage_aggs = [
        active_hours,
        uris,
        tabs_opened,
        windows_opened,
        ping_count
    ]

# agg fields: as health ping ----------------------------------

schema = BooleanType()
as_health_default_homepage_udf = udf(as_health_default_homepage, schema)

schema = BooleanType()
as_health_default_newtab_udf = udf(as_health_default_newtab, schema)


# note, we're just checking if these are set to true at ANY point in the day
# so if a user has it set to true, and then sets it to false, it'll show up at true
as_health_aggs = [
    F.max(
        as_health_default_homepage_udf(F.col('value'))
        ).alias('has_default_homepage'),
    F.max(
        as_health_default_newtab_udf(F.col('value'))
        ).alias('has_default_newtb'),
    F.count('activity_dt').alias('as_health_ping_count')
]

# agg fields: as session ping ----------------------------------

schema = BooleanType()
as_pref_setting_udf = udf(as_pref_setting, schema)

# note, we're just checking if these are set to true at ANY point in the day
# so if a user has it set to true, and then sets it to false, it'll show up at true
as_session_aggs = [
    F.max(
        as_pref_setting_udf(F.col('user_prefs'), F.lit(1))
        ).alias('has_search'),
    F.max(
        as_pref_setting_udf(F.col('user_prefs'), F.lit(2))
        ).alias('has_topsites'),
    F.max(
        as_pref_setting_udf(F.col('user_prefs'), F.lit(4))
        ).alias('has_pocket'),
    F.max(
        as_pref_setting_udf(F.col('user_prefs'), F.lit(8))
        ).alias('has_highlights'),
    F.max(
        as_pref_setting_udf(F.col('user_prefs'), F.lit(16))
        ).alias('has_snippets'),
    F.countDistinct('session_id').alias('as_sessions_count'),
    F.count('activity_dt').alias('as_session_ping_count')
]


# agg fields: as click ping ----------------------------------

TOPSITE_CONDITION = (F.col('source') == 'TOP_SITES')
HIGHLIGHT_CONDITION = (F.col('source') == 'HIGHLIGHTS')
HOMEPAGE_CONDITION = (F.col('page') == 'about:home')
NEWTAB_CONDITION = (F.col('page') == 'about:newtab')

# note: don't deal with nulls, if they didn't do it, they get a 0
# note: the generic click types can include about:welcome, but
#       the page specific ones don't
as_clicks_aggs = [
    # topsites
    F.sum(F.when(
            TOPSITE_CONDITION,
            1).otherwise(0)
          ).alias('topsite_clicks'),
    F.sum(F.when(
            TOPSITE_CONDITION & HOMEPAGE_CONDITION,
            1).otherwise(0)
          ).alias('topsite_homepage_clicks'),
    F.sum(F.when(
            TOPSITE_CONDITION & NEWTAB_CONDITION,
            1).otherwise(0)
          ).alias('topsite_newtab_clicks'),
    # highlights
    F.sum(F.when(
            HIGHLIGHT_CONDITION,
            1).otherwise(0)
          ).alias('highlight_clicks'),
    F.sum(F.when(
            HIGHLIGHT_CONDITION & HOMEPAGE_CONDITION,
            1).otherwise(0)
          ).alias('highlight_homepage_clicks'),
    F.sum(F.when(
            HIGHLIGHT_CONDITION & NEWTAB_CONDITION,
            1).otherwise(0)
          ).alias('highlight_newtab_clicks'),
    # combined
    F.sum(F.when(
            (TOPSITE_CONDITION | HIGHLIGHT_CONDITION),
            1).otherwise(0)
          ).alias('combined_clicks'),
    F.sum(F.when(
            (TOPSITE_CONDITION | HIGHLIGHT_CONDITION) & HOMEPAGE_CONDITION,
            1).otherwise(0)
          ).alias('combined_homepage_clicks'),
    F.sum(F.when(
            (TOPSITE_CONDITION | HIGHLIGHT_CONDITION) & NEWTAB_CONDITION,
            1).otherwise(0)
          ).alias('combined_newtab_clicks')
]


# agg fields: as snippet ping ----------------------------------

# note: don't deal with nulls or page, if they did it, they
# get a 1, otherwise 0
# note: unlike the other aggs, which are statically defined
#       this is defined as a function you have to run to
#       to produce the aggs list
#       this is because we'll want to parameterize based
#       on the message_id we are interested in
def get_as_snippets_aggs(message_id=None):
    """
    given a message_id for a snippet, return the agg functions
    for how many times that snippet was shown, blocked, dismissed,
    or clicked.

    if no message_id provided, then returns agg functions for all
    snippets.

    note: returns a list of agg functions for use in:
            daily_usage_df(df, agg_functions, client_fields=client_fields)
    """
    ID_CONDITION = F.lit(True)
    field_str = ''
    if message_id:
        ID_CONDITION = (F.col('message_id') == message_id)
        field_str = message_id

    IMPRESSION_CONDITION = (F.col('event') == 'IMPRESSION')
    BLOCK_CONDITION = (F.col('event') == 'BLOCK')
    DISMISS_CONDITION = (F.col('event') == 'DISMISS')
    CLICK_BUTTON_CONDITION = (F.col('event') == 'CLICK_BUTTON')

    as_snippets_aggs = [
        F.sum(F.when(
            (IMPRESSION_CONDITION & ID_CONDITION),
            1
        ).otherwise(0)
                 ).alias('snippet_impression_' + field_str),
        F.sum(F.when(
            (BLOCK_CONDITION & ID_CONDITION),
            1
        ).otherwise(0)
                 ).alias('snippet_block_' + field_str),
        F.sum(F.when(
            (DISMISS_CONDITION & ID_CONDITION),
            1
        ).otherwise(0)
                 ).alias('snippet_dismiss_' + field_str),
        F.sum(F.when(
            (CLICK_BUTTON_CONDITION & ID_CONDITION),
            1
        ).otherwise(0)
                 ).alias('snippet_click_' + field_str),
    ]

    return as_snippets_aggs


# ------------------- full client-daily df ------------------

# we always expect these columns in the df
# for use with joining
COMMON_COLUMNS = [
                 'client_id',
                 'branch',
                 'activity_dt',
                 'enrollment_dt',
                 'ping_count_enrollment',
                 'unenrollment_dt',
                 'ping_count_unenrollment'
                 ]


def null_safe_join(df1, df2, join_cols=COMMON_COLUMNS):
    """
    join dfs with the behavior that joining columns that
    are null will be treated as equal (so we don't get)
    multiple rows when the dfs have nulls in the same
    columns
    """

    def rename_col(df, column_name):
        df = df.withColumn(column_name + '_temp', F.col(column_name))
        df = df.drop(F.col(column_name))
        return df

    def null_safe_condition(column_name):
        return F.col(column_name).eqNullSafe(F.col(column_name + '_temp'))

    def get_null_safe_conditions(cols):
        expression = null_safe_condition(cols[0])
        for column_name in cols[1:]:
            expression = expression & null_safe_condition(column_name)
        return expression

    for column_name in join_cols:
        df2 = rename_col(df2, column_name)

    df = df1.join(df2, get_null_safe_conditions(join_cols), how='full')

    for column_name in join_cols:
        df = df.withColumn(column_name,
                           F.coalesce(F.col(column_name),
                                      F.col(column_name + '_temp')
                                      )
                           )    # not sure this is necessary...
        df = df.drop(F.col(column_name + '_temp'))

    return df


def overall_client(dfs, join_cols=COMMON_COLUMNS):
    """
    get multiple (client_id, branch, activity_dt) dfs and
    join into single df
    """

    base_df = dfs[0]

    for df in dfs[1:]:
        base_df = null_safe_join(base_df, df, join_cols)

    return base_df


def cleanup_no_activity_rows(df, activity_field='activity_dt'):
    """
    each df that was joined can produce "empty" activity rows for
    clients that didn't have activity in that df's activity
    this can blow up so you have "empty" rows for clients that
    did have activity in one of the dfs. this just cleans it up
    so that (client_id, branch) combos that have zero activity
    only get one row
    """

    df_has_activity = df.filter("{} is not null".format(activity_field))\
                        .select([
                            F.col('client_id').alias('client_id_temp'),
                            F.col('branch').alias('branch_temp')
                                ]).distinct()

    df = df.join(df_has_activity,
                 F.isnull(F.col('activity_dt'))
                 & (F.col('client_id') == F.col('client_id_temp'))
                 & (F.col('branch') == F.col('branch_temp')),
                 how='left'
                 )
    df = df.filter(F.isnull(F.col('client_id_temp')))
    df = df.drop('client_id_temp').drop('branch_temp')

    return df


NULLABLE_COLS = [
    'client_id',
    'branch',
    'activity_dt',
    'enrollment_dt',
    'ping_count_enrollment',
    'unenrollment_dt',
    'ping_count_unenrollment',
    'has_default_homepage',
    'has_default_newtb',
    'has_search',
    'has_topsites',
    'has_pocket',
    'has_highlights',
    'has_snippets'
]


def cleanup_nulls_df(df, nullable_cols=NULLABLE_COLS):
    """
    fill nulls with 0 for all columns except for the nullable cols
    """

    def clean_nulls(column, clean=True):
        if clean:
            return F.coalesce(F.col(column), F.lit(0)).alias(column)
        else:
            return F.col(column)

    cleaned_cols = []
    for col in df.columns:
        if col in nullable_cols:
            cleaned_cols.append(clean_nulls(col, False))
        else:
            cleaned_cols.append(clean_nulls(col, True))

    return df.select(cleaned_cols)



# activity stream utils

# get_tiles_querier(spark, dbutils)
# validate_as_data_quality(sql_query, dates, dbutils)


# activity stream utils: imports
from pyspark.sql import SparkSession
# local dependences
from .util import write_parquet_to_s3, read_parquet_from_s3, date_to_string
from .constants import S3_ROOT

# --------------- accessing activity stream data utils ---------------


spark = SparkSession.builder.getOrCreate()


def get_tiles_querier(spark, dbutils):
    """Return a function that queries the Tiles db.

    Example usage::

        query_tiles = get_tiles_querier(spark, dbutils)
        df = query_tiles('select 1;')
    """
    def query_tiles(sql_query):
        tempdir = 's3n://mozilla-databricks-telemetry-test/tiles-redshift/_temp'
        hostname = 'databricks-tiles-redshift.data.mozaws.net'
        port = 5432
        jdbc_url = 'jdbc:postgresql://{h}:{p}/tiles?user={u}&password={pw}'.format(
            h=hostname, p=port,
            u=dbutils.secrets.get('tiles-redshift', 'username'),
            pw=dbutils.secrets.get('tiles-redshift', 'password'),
        ) + '&ssl=true&sslMode=verify-ca'

        return spark.read.format(
            'com.databricks.spark.redshift'
        ).option(
            'forward_spark_s3_credentials', True
        ).option(
            'url', jdbc_url
        ).option(
            'tempdir', tempdir
        ).option(
            'query', sql_query
        ).load()

    return query_tiles


def validate_as_data_quality(sql_query, dates, dbutils):
    """
    for a given query and range/list of dates, check if
    databricks can pull the data into parquet at all
    most likely outcomes are:
        1) data pulls correctly (the counts are correct)
        2) spark can't communicate with the redshift db
           at which point you just wait / switch clusters /
           try again.
        3) pull fails and there will be a long py4 message
    """
    print("--------------------------------------------------")
    print("--------------------------------------------------")
    print("--------------------------------------------------")
    TEMP_DATA_DUMP = S3_ROOT + "activity-stream/temp-testing-dir.parquet"
    print("using {} as temporary directory".format(TEMP_DATA_DUMP))
    print("checking {} dates".format(str(dates)))
    for date in dates:
        if type(date) != str:
            date = date_to_string(date, format='%Y-%m-%d')
        q = sql_query.format(START_DT=date,
                             END_DT=date)
        print("\n\n\n")
        print("--------------------------------------------------")
        print(date)
        print("--------------------------------------------------")
        print("checking:")
        print(q)
        query_tiles = get_tiles_querier(spark, dbutils)
        try:
            test_data = query_tiles(q)
            test_count = test_data.count()

            write_parquet_to_s3(test_data, TEMP_DATA_DUMP)
            confirm_data = read_parquet_from_s3(TEMP_DATA_DUMP)

            confirm_count = confirm_data.count()
            if test_count == confirm_count:
                print("{}: data pulled and counts confirmed".format(date))
            elif test_count != confirm_count:
                print("{}: counts don't match").format(date)
        except Exception as E:
            print("{}: has error".format(date))
            print(E)


# --------------- parsing activity stream data utils ---------------

def as_experiment_field(shield_ids):
    """
    parse the shield_id field in activity stream data and return standard telemetry
    experiments field
    """
    try:
        if shield_ids and shield_ids != 'n/a':
            experiments = shield_ids.split(';')
            experiments = [exp for exp in experiments if exp != '']
            exp_dict = {}
            for i in experiments:
                exp_dict[i.split(':')[0]] = i.split(':')[1]
            if exp_dict != {}:
                return exp_dict
    except:  # noqa
        return


def as_pref_setting(user_prefs, setting):
    """
    parse as user_pref field (in sessions)
    and return condition for single setting
        setting is int for setting (see as docs)
    note - returns null if:
        1) null value
        2) value is less then 1
        3) value can't be coerced into integer
    """
    if type(user_prefs) == int and user_prefs >= 0:
        if user_prefs & setting == 0:
            return False
        if user_prefs & setting > 0:
            return True


def as_health_default_homepage(value):
    """
    parse value (in as health pings) to see if homepage
    is set to default
    note - returns null if:
        1) null value
    """
    if value:
        value = str(value)
        if value in ['0', '4', '8', '12']:
            return True
        else:
            return False


def as_health_default_newtab(value):
    """
    parse value (in as health pings) to see if newtab
    is set to default
    note - returns null if:
        1) null value
    """
    if value:
        value = str(value)
        if value in ['0', '1', '2', '3']:
            return True
        else:
            return False



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import pandas as pd
from datetime import timedelta

from dscontrib.jmccrosky.forecast.output import resetOuputTable, writeForecasts
from dscontrib.jmccrosky.forecast.data import getKPIData, getNondesktopData
from dscontrib.jmccrosky.forecast.data import getNondesktopNoFireData
from dscontrib.jmccrosky.forecast.utils import getLatestDate


_FIRST_MODEL_DATES = {
      'Fennec iOS': pd.to_datetime("2019-03-08").date(),
      'fxa_global': pd.to_datetime("2019-03-08").date(),
      'Firefox Lite': pd.to_datetime("2019-05-20").date(),
      'Focus iOS': pd.to_datetime("2019-03-08").date(),
      'Fenix': pd.to_datetime("2019-07-05").date(),
      'nondesktop_tier1': pd.to_datetime("2019-03-08").date(),
      'desktop_global': pd.to_datetime("2019-03-08").date(),
      'nondesktop_global': pd.to_datetime("2019-03-08").date(),
      'FirefoxConnect': pd.to_datetime("2019-03-08").date(),
      'FirefoxForFireTV': pd.to_datetime("2019-03-08").date(),
      'Fennec Android': pd.to_datetime("2019-03-08").date(),
      'desktop_tier1': pd.to_datetime("2019-03-08").date(),
      'fxa_tier1': pd.to_datetime("2019-03-08").date(),
      'Focus Android': pd.to_datetime("2019-03-08").date(),
      'nondesktop_nofire_global': pd.to_datetime("2019-03-08").date(),
      'nondesktop_nofire_tier1': pd.to_datetime("2019-03-08").date(),
}
_FORECAST_HORIZON = pd.to_datetime("2019-12-31").date()
_BQ_PROJECT = "moz-fx-data-derived-datasets"
_BQ_DATASET = "analysis"
_BQ_TABLE = "jmccrosky_test"


def updateTable(bqClient):
    kpiData = getKPIData(bqClient)
    nondesktopData = getNondesktopData(bqClient)
    nondesktopnofireData = getNondesktopNoFireData(bqClient)
    data = kpiData
    data.update(nondesktopData)
    data.update(nondesktopnofireData)
    dataset = bqClient.dataset(_BQ_DATASET)
    tableref = dataset.table(_BQ_TABLE)
    table = bqClient.get_table(tableref)
    for product in data.keys():
        latestDate = getLatestDate(
            bqClient, _BQ_PROJECT, _BQ_DATASET, _BQ_TABLE, product, "asofdate"
        )
        if latestDate is not None:
            startDate = latestDate + timedelta(days=1)
        else:
            startDate = _FIRST_MODEL_DATES[product]
        model_dates = pd.date_range(
            startDate,
            data[product].ds.max() - timedelta(days=1)
        )
        for model_date in model_dates:
            writeForecasts(
                bqClient, table, model_date.date(),
                _FORECAST_HORIZON, data[product], product
            )


def replaceTable(bqClient):
    kpiData = getKPIData(bqClient)
    nondesktopData = getNondesktopData(bqClient)
    nondesktopnofireData = getNondesktopNoFireData(bqClient)
    data = kpiData
    data.update(nondesktopData)
    data.update(nondesktopnofireData)
    table = resetOuputTable(bqClient, _BQ_PROJECT, _BQ_DATASET, _BQ_TABLE)
    for product in data.keys():
        model_dates = pd.date_range(
            _FIRST_MODEL_DATES[product],
            data[product].ds.max() - timedelta(days=1)
        )
        for model_date in model_dates:
            writeForecasts(
                bqClient, table, model_date.date(),
                _FORECAST_HORIZON, data[product], product
            )



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

import mozanalysis.bayesian_stats.binary as mabin
import mozanalysis.bayesian_stats.bayesian_bootstrap as mabb
import mozanalysis.bayesian_stats.survival_func as masf


def plot_ts(t_df, col_label, stats_model, ref_branch_label='control', sc=None):
    """Plot time series data for a metric.

    Args:
        t_df (dict): A dictionary keyed by time (e.g. an int representing
            the number of days since enrollment), where the values are
            standard format dfs of queried experiment data for that time.
    """
    # Check the number of enrollments is equal for all parts of time series
    assert len({len(v) for v in t_df.values()})

    data = crunch_nums_ts(t_df, col_label, stats_model, ref_branch_label, sc=sc)

    fig, (ax1, ax2) = plt.subplots(nrows=2, sharex=True, figsize=(6, 10))

    # Hackily guess whether to interpolate the data points
    if len(data['individual']) >= 10:
        plot_means_line(ax1, data['individual'], ref_branch_label)
        plot_uplifts_line(ax2, data['comparative'], ref_branch_label)
    else:
        plot_means_scatter(ax1, data['individual'])
        plot_uplifts_scatter(ax2, data['comparative'])

    ax1.set_title("{} per user over time".format(col_label))
    ax1.legend()
    ax1.set_ylabel(col_label)

    ax2.set_xlabel('Days since enrollment')
    ax2.set_ylabel('Uplift relative to {}'.format(ref_branch_label))
    fig.tight_layout()
    return fig


def plot_survival(df, col_label, ref_branch_label='control', thresholds=None):
    data = masf.compare_branches(df, col_label, ref_branch_label, thresholds)

    fig, (ax1, ax2) = plt.subplots(nrows=2, sharex=True, figsize=(6, 10))

    for ax in (ax1, ax2):
        ax.set_prop_cycle('color', plt.cm.tab20b(np.linspace(0, 1, 5)))

    plot_means_line(ax1, data['individual'], ref_branch_label)
    plot_uplifts_line(ax2, data['comparative'])

    ax1.set_title(
        'Survival fn: Fraction of users with {} > x'.format(col_label)
    )
    ax1.set_ylim((0, 1))
    ax1.set_ylabel('Fraction of users')
    ax1.legend()
    ax2.set_xlabel(col_label)
    ax2.set_ylabel('Uplift relative to {}'.format(ref_branch_label))

    fig.tight_layout()
    return fig


def plot_means_line(ax, branch_x_stats, ref_branch_label='control'):
    for branch_label in sort_branch_list(branch_x_stats.keys(), ref_branch_label):
        _plot_means_line(ax, branch_x_stats[branch_label], branch_label)


def plot_means_scatter(ax, branch_x_stats, ref_branch_label='control'):
    num_branches = len(branch_x_stats)

    for i, branch_label in enumerate(
        sort_branch_list(branch_x_stats.keys(), ref_branch_label)
    ):
        offset = i * 0.15 - (num_branches - 1) / 2 * 0.15
        _plot_means_scatter(
            ax, pd.DataFrame(branch_x_stats[branch_label]).T, branch_label, offset
        )


def _plot_means_line(ax, df, branch_label):
    line = ax.plot(df.index, df['mean'], label=branch_label)[0]
    col = line.get_color()
    ax.fill_between(
        df.index,
        df['0.005'],
        df['0.995'],
        color=col,
        alpha=0.05
    )
    ax.fill_between(
        df.index,
        df['0.025'],
        df['0.975'],
        color=col,
        alpha=0.05
    )


def _plot_means_scatter(ax, df, branch_label, offset):
    # TODO: add an x offset between branches, and caps for matplotlib 3
    yerr_inner = (df[['0.025', '0.975']].T - df['mean']).abs().values
    yerr_outer = (df[['0.005', '0.995']].T - df['mean']).abs().values
    line = ax.errorbar(
        df.index + offset, df['mean'], yerr=yerr_inner,
        fmt='.', elinewidth=3, capsize=0,
        label=branch_label
    )[0]
    col = line.get_color()
    ax.errorbar(
        df.index + offset, df['mean'], yerr=yerr_outer,
        fmt='.', capsize=3, color=col, ecolor=col, label=None
    )


def plot_uplifts_line(ax, branch_x_stats):
    for branch_label in sorted(branch_x_stats.keys()):
        _plot_uplifts_line(ax, branch_x_stats[branch_label])

    xmin, xmax = ax.get_xlim()
    ax.plot(
        [0, xmin, xmax],
        [0, 0, 0],
        'k--', label='zero'
    )


def _plot_uplifts_line(ax, df):
    line = ax.plot(df.index, df[('rel_uplift', 'exp')])[0]
    col = line.get_color()
    ax.fill_between(
        df.index,
        df[('rel_uplift', '0.005')],
        df[('rel_uplift', '0.995')],
        color=col,
        alpha=0.05
    )
    ax.fill_between(
        df.index,
        df[('rel_uplift', '0.025')],
        df[('rel_uplift', '0.975')],
        color=col,
        alpha=0.05
    )


def plot_uplifts_scatter(ax, branch_x_stats):
    for branch_label in sorted(branch_x_stats.keys()):
        _plot_uplifts_scatter(ax, branch_x_stats[branch_label])

    xmin, xmax = ax.get_xlim()
    ax.plot(
        [0, xmin - 1, xmax + 1],
        [0, 0, 0],
        'k--', label='zero'
    )


def _plot_uplifts_scatter(ax, x_df):
    df = pd.DataFrame(x_df, columns=sorted(x_df.keys())).T
    yerr_inner = (
        df[[('rel_uplift', '0.025'), ('rel_uplift', '0.975')]].T
        - df[('rel_uplift', 'exp')]
    ).abs().values
    yerr_outer = (
        df[[('rel_uplift', '0.005'), ('rel_uplift', '0.995')]].T
        - df[('rel_uplift', 'exp')]
    ).abs().values
    line = ax.errorbar(
        df.index, df[('rel_uplift', 'exp')], yerr=yerr_inner,
        fmt='.', elinewidth=3, capsize=0,
    )[0]
    col = line.get_color()
    ax.errorbar(
        df.index, df[('rel_uplift', 'exp')], yerr=yerr_outer,
        fmt='.', capsize=3, color=col, ecolor=col, label=None
    )
    # # matplotlib 1 :(
    # ax.set_xticks(df.index)
    # ax.set_xlim((min(df.index) - 1, max(df.index) + 1))


def crunch_nums_ts(ts, col_label, stats_model, ref_branch_label='control', sc=None):
    assert all_eq((len(v) for v in ts.values()))
    assert all_eq((set(tuple(sorted(v.branch.unique())) for v in ts.values())))

    branch_list = next(iter(ts.values())).branch.unique()
    # # Maybe defaultdicts are offensive because they hide the schema?
    # res = collections.defaultdict(lambda: collections.defaultdict(dict))
    res = {
        'comparative': {
            b: {
                t: None for t in ts.keys()
            } for b in branch_list if b != ref_branch_label
        },
        'individual': {
            b: {
                t: None for t in ts.keys()
            } for b in branch_list
        },
    }

    # TODO: this really smells like a map then a zip?
    for k, v in ts.items():
        if stats_model == 'beta':
            bla = mabin.compare_branches(
                v, col_label, ref_branch_label=ref_branch_label
            )
        elif stats_model == 'bootstrap':
            assert sc is not None
            bla = mabb.compare_branches(
                sc, v, col_label,
                ref_branch_label=ref_branch_label, threshold_quantile=0.9999
            )
        else:
            raise NotImplementedError

        for branch, data in bla['comparative'].items():
            res['comparative'][branch][k] = data
        for branch, data in bla['individual'].items():
            res['individual'][branch][k] = data

    return res


def sort_branch_list(branch_labels, ref_branch_label='control'):
    """Order branch labels for plotting, consistently.

    The control branch goes last because some plots include a series
    for it, and some plots don't, and we want the colors to be
    consistent between these plots.
    """
    assert ref_branch_label in branch_labels

    return sorted(set(branch_labels) - {ref_branch_label}) + [ref_branch_label]


def all_eq(iterable):
    return len(set(iterable)) <= 1



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from pyspark.sql import functions as F
from pyspark.sql.functions import col, lit
from pyspark.sql import Window

# Source table
_TABLE_SOURCE = "clients_daily"

# Source Columns
_COL_ID = col("client_id")
_COL_DATE = col("submission_date_s3")
_COL_URI_COUNT = F.col("scalar_parent_browser_engagement_total_uri_count_sum")
_COL_PC_DATE = col("profile_creation_date")
_COL_SAMPLE = col("sample_id")

# Other constants
_NUM_ADAU_THRESHOLD = 5
_DATE_PARSED = F.to_date(_COL_DATE, 'yyyyMMdd')
_MAP_NATURAL_DIMENSIONS = {
    "country": "country",
    "channel": "channel",
    "os": "os",
    "os_version": "os_version",
    "normalized_channel": "normalized_channel",
    "active_hours_sum": "active_hours_sum",
    "search_count_all": "search_count_all",
    "distribution_id": "distribution_id",
}

# Data file columns (TODO)


def createDataFile(
    start_date, end_date, spark_instance, jackknife_buckets, sample_percent, output_path
):
    feature_data_phase1 = spark_instance.table(_TABLE_SOURCE).select(
        [
            _COL_ID.alias("id"),
            _DATE_PARSED.alias("date"),
            # TODO: Use MD5 instead of CRC32
            (F.floor(F.crc32(_COL_ID) / 100) % jackknife_buckets).alias("bucket"),
            lit(1).alias("is_active"),
            F.when(
                _COL_URI_COUNT >= _NUM_ADAU_THRESHOLD, 1
            ).otherwise(
                0
            ).alias("is_active_active"),
            F.to_date(_COL_PC_DATE).alias("profile_creation_date")
        ] + list(_MAP_NATURAL_DIMENSIONS.keys())
    ).filter(
        (_DATE_PARSED.between(start_date, end_date)) &
        (_COL_SAMPLE < sample_percent)
    ).withColumn(
        "young_profile",
        F.when(
            col("date") < F.date_add(col("profile_creation_date"), 14), "TRUE"
        ).otherwise(
            "FALSE"
        )
    )

    new_profile_window = Window.partitionBy(col("id")).orderBy(col("date"))
    new_profile_data = feature_data_phase1.filter(
        (col("date") >= col("profile_creation_date")) &
        (col("date") <= F.date_add(col("profile_creation_date"), 6))
    ).select(
        "*",
        F.rank().over(new_profile_window).alias('rank')
    ).filter(
        col('rank') == 1
    ).withColumn(
        "new_profile", lit(1)
    ).drop(
        "date"
    ).withColumn(
        "date", col("profile_creation_date")
    )

    feature_data = feature_data_phase1.alias("fd").join(
        new_profile_data.alias("np"),
        (col("fd.id") == col("np.id")) & (col("fd.date") == col("np.date")),
        how='full',
    ).select(
        [F.coalesce(col("np.new_profile"), lit(0)).alias("new_profile")] +
        [F.coalesce(col("fd.is_active"), lit(0)).alias("is_active")] +
        [F.coalesce(col("fd.is_active_active"), lit(0)).alias("is_active_active")] +
        [
            F.coalesce(col("fd.{}".format(c)), col("np.{}".format(c))).alias(c)
            for c in feature_data_phase1.columns
            if c not in ["is_active", "is_active_active"]
        ]
    )

    once_ever_profiles = feature_data.filter(col("is_active") == 1).groupBy(
        "id"
    ).count(
    ).filter(
        col("count") == 1
    ).select(
        "id"
    ).withColumn(
        "single_day_profile", lit("1")
    )

    feature_data = feature_data.alias("fd").join(
        once_ever_profiles.alias("oep"),
        "id",
        "outer"
    ).fillna(
        {"single_day_profile": "0"}
    )

    ppi_profiles = spark_instance.table("main_summary").select(
      col("client_id").alias("id"),
      lit(1).alias("ppi")
    ).filter(
        '''submission_date_s3 >= '20190121'
        AND scalar_parent_startup_profile_selection_reason IN (
            'firstrun-skipped-default', 'restart-skipped-default'
        )'''
    )

    feature_data = feature_data.alias("fd").join(
        ppi_profiles.alias("ppip"),
        "id",
        "left"
    ).fillna(
        {"ppi": 0}
    )

    feature_data.write.partitionBy("date").mode('overwrite').parquet(output_path)


def readDataFile(spark_instance, data_path):
    return spark_instance.read.parquet(data_path)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import numpy as np
import pandas as pd
import plotly.graph_objs as go
from datetime import timedelta


# Calculate Mean Absolute Percentage Error of forecast
def calcMAPE(true, predicted):
    mask = true != 0
    return (np.fabs(true - predicted)/true)[mask].mean() * 100


# Calculate Mean Relative Error of forecast
def calcMRE(true, predicted):
    mask = true != 0
    return ((true - predicted)/true)[mask].mean() * 100


def calcLogRatio(true, predicted):
    logratio = np.mean(np.log(predicted) - np.log(true))
    return logratio


# Get most recent date in table
def getLatestDate(bqClient, project, dataset, table, product, field):
    query = '''
        SELECT
            MAX({field}) as date
        FROM
            `{project}.{dataset}.{table}`
        WHERE
            datasource="{product}"
    '''.format(
        project=project, dataset=dataset, table=table, field=field, product=product
    )
    data = bqClient.query(query).to_dataframe()
    if len(data) == 0:
        return None
    return bqClient.query(query).to_dataframe()['date'][0]


def splitData(data, firstTrainDate, firstHoldoutDate, firstTestDate, lastTestDate):
    temp = data.set_index('ds')
    split_data = {
        "training": temp[
            firstTrainDate:(firstHoldoutDate - timedelta(days=1))
        ].reset_index(),
        "holdout": temp[
            firstHoldoutDate:(firstTestDate - timedelta(days=1))
        ].reset_index(),
        "test": temp[firstTestDate:lastTestDate].reset_index(),
        "all": temp.reset_index(),
    }
    return split_data


def s2d(stringDate):
    return pd.to_datetime(stringDate).date()


def matchDates(data, forecast):
    temp = forecast.copy()
    temp['ds'] = temp.ds.dt.date
    return data.merge(temp, on="ds", how="inner")


def getLayout(title, xaxis, yaxis):
    return go.Layout(
        title=title,
        xaxis=go.layout.XAxis(
            title=go.layout.xaxis.Title(
                text=xaxis
            )
        ),
        yaxis=go.layout.YAxis(
            title=go.layout.yaxis.Title(
                text=yaxis
            )
        ),
    )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


usage_criteria = {
  "Any Activity": "is_active",
  "5 URI": "is_active_active",
  "New Profile": "new_profile",
  "New Profile Requiring Ping on Same Day": "new_profile_old",
}



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import pandas as pd
import numpy as np
from datetime import timedelta
from collections import defaultdict
from plotly.offline import plot
import plotly.graph_objs as go
from dscontrib.jmccrosky.forecast.utils import s2d, matchDates, getLayout


def _getSinglePrediciton(model, data, trainingEndDate, targetDate):
    model.fit(data.query("ds <= @trainingEndDate"))
    forecast_period = pd.DataFrame({'ds': [s2d(targetDate)]})
    forecast = model.predict(forecast_period)
    return (forecast.yhat[0], forecast.yhat_lower[0], forecast.yhat_upper[0])


def ValidateStability(modelGen, data, trainingEndDateRange, targetDate):
    dates = []
    values = []
    for d in trainingEndDateRange:
        predictions = []
        predictions.append(_getSinglePrediciton(modelGen(), data, d, targetDate))
        values.append(predictions)
        dates.append(d)
    data = pd.DataFrame({
        "date": dates,
        "Predicted MAU for {}".format(targetDate): [i[0][0] for i in values],
        "upper": [i[0][1] for i in values],
        "lower": [i[0][2] for i in values],
    })
    return plot(
        {
            "data": [
                go.Scatter(
                    x=data['date'],
                    y=data["Predicted MAU for {}".format(targetDate)],
                    name="Prediction",
                ),
                go.Scatter(
                    x=data['date'],
                    y=data['upper'],
                    fill='tonexty',
                    mode='none',
                    name='upper 80% CI',
                ),
                go.Scatter(
                    x=data['date'],
                    y=data['lower'],
                    fill='tonexty',
                    mode='none',
                    name='lower 80% CI',
                ),
            ],
            "layout": getLayout(
                ("Predictions of MAU for {} using model "
                 "fit on data up to each date").format(targetDate),
                "Model Trained On Data Up To Date",
                "Prediction for {}".format(targetDate)
            )
        },
        output_type="div"
    )


def _getMetricForRange(model, data, trainingEndDate, metric):
    forecastStart = trainingEndDate + timedelta(days=1)
    forecastEnd = data.ds.max()
    model.fit(data.query("ds <= @trainingEndDate"))
    forecast_period = pd.DataFrame({'ds': pd.date_range(forecastStart, forecastEnd)})
    forecast = model.predict(forecast_period)
    matched = matchDates(
        data.query("ds <= @forecastEnd & ds >= @forecastStart"),
        forecast
    )
    metric = metric(
        np.array(matched.y),
        np.array(matched.yhat),
    )
    return metric


def ValidateMetric(modelGen, data, trainingEndDateRange, metric, metricName):
    dates = []
    mapes = []
    for d in trainingEndDateRange:
        mapes.append(_getMetricForRange(modelGen(), data, d, metric))
        dates.append(d)
    data = pd.DataFrame({"date": dates, metricName: mapes})
    return plot(
        {
            "data": [go.Scatter(x=data['date'], y=data[metricName], name=metricName)],
            "layout": getLayout(
                "{} for model trained up to date".format(metricName),
                "Model Trained On Data Up To Date",
                "{} Metric Value".format(metricName)
            )
        },
        output_type="div"
    )


def _getMetricTrace(model, data, trainingEndDate, metric, metricName):
    forecastStart = trainingEndDate + timedelta(days=1)
    forecastEnd = data.ds.max()
    model.fit(data.query("ds <= @trainingEndDate"))
    forecastPeriod = pd.DataFrame({'ds': pd.date_range(forecastStart, forecastEnd)})
    forecast = model.predict(forecastPeriod)
    matched = matchDates(
        data,
        forecast
    )
    return pd.DataFrame({
        "ds": matched.ds,
        metricName: [
            metric(
                np.array(matched.query("ds == @d").y),
                np.array(matched.query("ds == @d").yhat)
            )
            for d in matched.ds
        ]
    })


def ValidateTraces(modelGen, data, trainingEndDateRange, metric, metricName):
    traces = []
    for d in trainingEndDateRange:
        traces.append(_getMetricTrace(modelGen(), data, d, metric, metricName))
    return plot(
        {
            "data":
                [
                    go.Scatter(x=d['ds'], y=d[metricName])
                    for d in traces
                ],
                "layout": getLayout(
                    "Model Traces of {}".format(metricName),
                    "Prediction Date",
                    "{} Metric Value".format(metricName)
                )
        },
        output_type="div",
    )


def _accumulateHorizonMetrics(model, data, trainingEndDate, metric, metricValues):
    forecastStart = trainingEndDate + timedelta(days=1)
    forecastEnd = data.ds.max()
    model.fit(data.query("ds <= @trainingEndDate"))
    forecastPeriod = pd.DataFrame({'ds': pd.date_range(forecastStart, forecastEnd)})
    forecast = model.predict(forecastPeriod)
    matched = matchDates(
        data,
        forecast
    )
    for d in matched.ds:
        metricValue = metric(
            np.array(matched.query("ds == @d").y),
            np.array(matched.query("ds == @d").yhat)
        )
        horizon = (d - trainingEndDate).days
        metricValues[horizon].append(metricValue)


def ValidateMetricHorizon(modelGen, data, trainingEndDateRange, metric, metricName):
    metricValues = defaultdict(lambda: [])
    for d in trainingEndDateRange:
        _accumulateHorizonMetrics(modelGen(), data, d, metric, metricValues)
    data = pd.DataFrame({
        "horizon": [i for i in metricValues.keys()],
        metricName: [np.mean(metricValues[h]) for h in metricValues.keys()]
    })
    return plot({
        "data": [
            go.Scatter(x=data['horizon'], y=data[metricName], name="metricName"),
        ],
        "layout": getLayout(
            "{} Metric By Model Horizon".format(metricName),
            "Model Horizon",
            "Mean {} Metric Value".format(metricName)
        )
    }, output_type="div")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import pandas as pd
import numpy as np

from pyspark.sql import functions as F
from pyspark.sql import Window
from astropy.stats import jackknife_stats


def calculateDateWindow(
        plot_start_date, plot_end_date, smoothing, comparison_mode, pre_days, post_days
):
    start_window = plot_start_date - \
        pd.DateOffset(days=smoothing-1) - \
        pd.DateOffset(days=pre_days) - \
        pd.to_timedelta('1 second')           # Needed due to pyspark between() bug
    end_window = plot_end_date + \
        pd.to_timedelta('1 second') + \
        pd.DateOffset(days=post_days)
    window = [(start_window, end_window)]
    if comparison_mode in ["YoY", "Last Year"]:
        window = window + [(
            start_window - pd.DateOffset(years=1),
            end_window - pd.DateOffset(years=1)
        )]
    return window


def doSmoothing(data, usage_criteria, dimension_cols, smoothing_lookback):
    windowSpec = Window.partitionBy(
        [data.bucket] + dimension_cols
    ).orderBy(
        data.date
    ).rowsBetween(
        -smoothing_lookback, 0
    )

    return data.withColumn(
        "n_", F.mean(data[usage_criteria]).over(windowSpec)
    ).drop(
        usage_criteria
    ).withColumnRenamed(
        "n_", usage_criteria
    )


def jackknifeCountCI(data, string_mode=False):
    count = np.sum(data)
    jackknife_buckets = len(data)
    estimate, bias, stderr, conf_interval = \
        jackknife_stats(np.array(data), np.mean, 0.95)
    if string_mode:
        # TODO: extract a CI formatting function with configurable decimal places
        return "({:.0f} - {:.0f})".format(
            conf_interval[0] * jackknife_buckets, conf_interval[1] * jackknife_buckets
        )
    else:
        return [
            (count - conf_interval[0] * jackknife_buckets),
            (conf_interval[1] * jackknife_buckets - count)
        ]


def jackknifeMeanCI(data, string_mode=False):
    mean = np.mean(data)
    estimate, bias, stderr, conf_interval = \
        jackknife_stats(np.array(data), np.mean, 0.95)
    if string_mode:
        return "({:.0f} - {:.0f})".format(conf_interval[0], conf_interval[1])
    else:
        return [mean - conf_interval[0], conf_interval[1] - mean]


def getPandasDimensionQuery(dimensions):
    if len(dimensions) == 0:
        return ""
    return " and " + " and ".join(["{}=='{}'".format(
        d, dimensions[d]) for d in dimensions]
    )


def dimensionName(dimension):
    return "|".join(["{}".format(dimension[d]) for d in dimension])


def longDimensionName(dimension):
    if len(dimension) > 0:
        return "|".join(["{}: {}".format(d, dimension[d]) for d in dimension])
    else:
        return "ALL"
